"""Tests the STACS Scanner Rule module."""

import os
import unittest

import stacs.scan


class STACSScannerRuleTestCase(unittest.TestCase):
    """Tests the STACS Scanner Rule module."""

    def setUp(self):
        """Ensure the application is setup for testing."""
        self.fixtures_path = os.path.join(
            os.path.dirname(os.path.abspath(__file__)), "fixtures/findings/"
        )

    def tearDown(self):
        """Ensure everything is torn down between tests."""
        pass

    def test_generate_sample(self):
        """Ensures that samples are correctly generated."""
        reduced_after_finding = stacs.scan.model.manifest.Entry(
            path=os.path.join(self.fixtures_path, "001.txt")
        )
        reduced_before_finding = stacs.scan.model.manifest.Entry(
            path=os.path.join(self.fixtures_path, "002.txt")
        )
        only_finding = stacs.scan.model.manifest.Entry(
            path=os.path.join(self.fixtures_path, "003.txt")
        )
        sufficent_before_after_finding = stacs.scan.model.manifest.Entry(
            path=os.path.join(self.fixtures_path, "004.txt")
        )

        # Check that the correct number of bytes were extracted before and after the
        # respective findings.
        context = stacs.scan.scanner.rules.generate_sample(
            reduced_after_finding,
            191,  # Offset.
            40,  # Size.
        )
        self.assertEqual(len(context.before), 20)
        self.assertEqual(len(context.finding), 40)
        self.assertEqual(len(context.after), 1)

        context = stacs.scan.scanner.rules.generate_sample(
            reduced_before_finding,
            3,  # Offset.
            40,  # Size.
        )
        self.assertEqual(len(context.before), 3)
        self.assertEqual(len(context.finding), 40)
        self.assertEqual(len(context.after), 20)

        context = stacs.scan.scanner.rules.generate_sample(
            only_finding,
            0,  # Offset.
            40,  # Size.
        )
        self.assertEqual(len(context.before), 0)
        self.assertEqual(len(context.finding), 40)
        self.assertEqual(len(context.after), 0)

        context = stacs.scan.scanner.rules.generate_sample(
            sufficent_before_after_finding,
            137,  # Offset.
            40,  # Size.
        )
        self.assertEqual(len(context.before), 20)
        self.assertEqual(len(context.finding), 40)
        self.assertEqual(len(context.after), 20)



"""Tests the STACS filepath loader."""

import os
import unittest


class STACSLoaderFilepathTestCase(unittest.TestCase):
    """Tests the STACS filepath loader."""

    def setUp(self):
        """Ensure the application is setup for testing."""
        self.fixtures_path = os.path.join(
            os.path.dirname(os.path.abspath(__file__)), "fixtures/"
        )

    def tearDown(self):
        """Ensure everything is torn down between tests."""
        pass



"""Provides an Apple Disk Image (DMG) parser and extractor.

SPDX-License-Identifier: BSD-3-Clause
"""

import bz2
import lzma
import os
import plistlib
import struct
import zlib
from collections import namedtuple
from typing import List

from pydantic import BaseModel, Extra, Field
from stacs.scan.exceptions import FileAccessException, InvalidFileException

# Structures names and geometry are via "Demystifying the DMG File Format"
# by Jonathan Levin (http://newosxbook.com/).
DMG_HEADER_MAGIC = b"koly"
DMG_HEADER = ">4sIIIQQQQQII16sII128sQQ120sII128sIQIII"
DMG_HEADER_MAGIC_SZ = len(DMG_HEADER_MAGIC)
DMG_HEADER_SZ = struct.calcsize(DMG_HEADER)

DMG_BLOCK_TABLE_MAGIC = b"mish"
DMG_BLOCK_TABLE = ">4sIQQQIIIIIIIIII128sI"
DMG_BLOCK_TABLE_MAGIC_SZ = len(DMG_BLOCK_TABLE_MAGIC)
DMG_BLOCK_TABLE_SZ = struct.calcsize(DMG_BLOCK_TABLE)

DMG_BLOCK_CHUNK = ">I4sQQQQ"
DMG_BLOCK_CHUNK_SZ = struct.calcsize(DMG_BLOCK_CHUNK)

DMGHeader = namedtuple(
    "DMGHeader",
    [
        "signature",
        "version",
        "header_size",
        "flags",
        "running_data_fork_offset",
        "data_fork_offset",
        "data_fork_length",
        "rsrc_fork_offset",
        "rsrc_fork_length",
        "segment_number",
        "segment_count",
        "segment_id",
        "data_checksum_type",
        "data_checksum_size",
        "data_checksum",
        "xml_offset",
        "xml_length",
        "reserved_1",
        "checksum_Type",
        "checksum_Size",
        "checksum",
        "image_variant",
        "sector_count",
        "reserved_2",
        "reserved_3",
        "reserved_4",
    ],
)
DMGBlockTable = namedtuple(
    "DMGBlockTable",
    [
        "signature",
        "version",
        "sector_number",
        "sector_count",
        "data_offset",
        "buffers_needed",
        "block_descriptors",
        "reserved_1",
        "reserved_2",
        "reserved_3",
        "reserved_4",
        "reserved_5",
        "reserved_6",
        "checksum_ype",
        "checksum_ize",
        "checksum",
        "chunk_count",
    ],
)
DMGBlockChunk = namedtuple(
    "DMGBlockChunk",
    [
        "type",
        "comment",
        "sector_number",
        "sector_count",
        "compressed_offset",
        "compressed_length",
    ],
)


class DMGBlock(BaseModel, extra=Extra.forbid):
    """Expresses a DMG block entry and its chunks."""

    name: str
    chunks: List[DMGBlockChunk] = Field([])


class DMG:
    """Provides an Apple Disk Image (DMG) parser and extractor."""

    def __init__(self, filepath: str):
        self.archive = filepath

        try:
            with open(self.archive, "rb") as fin:
                # DMG metadata is at the end of the file.
                fin.seek(-DMG_HEADER_SZ, 2)

                # Ensure the provided file is actually a DMG.
                if fin.read(DMG_HEADER_MAGIC_SZ) != DMG_HEADER_MAGIC:
                    raise InvalidFileException("File does not appear to be a DMG")

                # Rewind and attempt to read in header.
                fin.seek(-DMG_HEADER_MAGIC_SZ, 1)
                self._header = DMGHeader._make(
                    struct.unpack(DMG_HEADER, fin.read(DMG_HEADER_SZ))
                )

                # Read the XML property list.
                fin.seek(self._header.xml_offset, 0)
                self._plist = plistlib.loads(fin.read(self._header.xml_length))
        except OSError as err:
            raise FileAccessException(f"Unable to read archive: {err}")

    def _parse_blocks(self) -> List[DMGBlock]:
        """Recursively parse blocks and their associated chunks."""
        candidates = []

        # Read the BLKX entries from the resource-fork section of the plist.
        for entry in self._plist.get("resource-fork", {}).get("blkx", []):
            data = entry.get("Data")
            name = entry.get("Name")

            block = DMGBlock(name=name)
            table = DMGBlockTable._make(
                struct.unpack(DMG_BLOCK_TABLE, data[0:DMG_BLOCK_TABLE_SZ])
            )

            # Extract all blocks and their associated chunks from the encoded "Data"
            # inside of the extracted plist.
            start = DMG_BLOCK_TABLE_SZ

            for _ in range(0, table.chunk_count):
                end = start + DMG_BLOCK_CHUNK_SZ
                block.chunks.append(
                    DMGBlockChunk._make(struct.unpack(DMG_BLOCK_CHUNK, data[start:end]))
                )
                start = end

            candidates.append(block)

        return candidates

    def extract(self, destination):
        """Extract all blocks from the DMG to the optional destination directory."""
        parent = os.path.basename(self.archive)

        try:
            os.makedirs(destination, exist_ok=True)
        except OSError as err:
            raise FileAccessException(
                f"Unable to create directory during extraction: {err}"
            )

        # Process each chunk inside of each block. A DMG has multiple blocks, and a
        # block has N chunks.
        for idx, block in enumerate(self._parse_blocks()):
            output = os.path.join(destination, f"{parent}.{idx}.blob")

            for chunk in block.chunks:
                # Skip Ignored, Comment, and Last blocks (respectively).
                if chunk.type in [0x00000002, 0x7FFFFFFE, 0xFFFFFFFF]:
                    continue

                try:
                    with open(self.archive, "rb") as fin, open(output, "ab") as fout:
                        fin.seek(chunk.compressed_offset)

                        # 0x80000005 - Zlib.
                        if chunk.type == 0x80000005:
                            fout.write(
                                zlib.decompress(fin.read(chunk.compressed_length))
                            )

                        # 0x80000005 - BZ2.
                        if chunk.type == 0x80000006:
                            fout.write(
                                bz2.decompress(fin.read(chunk.compressed_length))
                            )

                        # 0x80000005 - LZMA.
                        if chunk.type == 0x80000008:
                            fout.write(
                                lzma.decompress(fin.read(chunk.compressed_length))
                            )

                        # 0x00000000 - Zero Fill.
                        if chunk.type == 0x00000000:
                            fout.write(b"\x00" * chunk.compressed_length)
                            continue
                except (OSError, lzma.LZMAError, ValueError) as err:
                    raise InvalidFileException(err)



"""Provides an eXtensible ARchive parser and extrator.

SPDX-License-Identifier: BSD-3-Clause
"""

import os
import struct
import xml.etree.ElementTree as ET
import zlib
from collections import namedtuple
from typing import List

from stacs.scan.constants import CHUNK_SIZE
from stacs.scan.exceptions import FileAccessException, InvalidFileException

XAR_MAGIC = b"xar!"
XAR_HEADER = ">4sHHQQI"
XAR_HEADER_SZ = struct.calcsize(XAR_HEADER)

# via xar/include/xar.h.in
XARHeader = namedtuple(
    "XARHeader",
    [
        "magic",
        "size",
        "version",
        "toc_length_compressed",
        "toc_length_uncompressed",
        "cksum_alg",
    ],
)

XAREntry = namedtuple(
    "XAREntry",
    [
        "length",
        "offset",
        "size",
        "encoding",
        "archived_cksum_kind",
        "archived_cksum",
        "path",
        "name",
        "kind",
    ],
)


class XAR:
    """Provides an eXtensible ARchive Format parser and extrator."""

    def __init__(self, filepath: str):
        self.archive = filepath

        try:
            with open(self.archive, "rb") as fin:
                # Ensure the provided file is actually a XAR.
                if fin.read(4) != XAR_MAGIC:
                    raise InvalidFileException("File does not appear to be a XAR")

                # Rewind and attempt to read in header.
                fin.seek(0)
                self._header = XARHeader._make(
                    struct.unpack(XAR_HEADER, fin.read(XAR_HEADER_SZ))
                )

                # Read and decompress the table-of-contents.
                fin.seek(self._header.size)

                self._toc = ET.fromstring(
                    str(
                        zlib.decompress(fin.read(self._header.toc_length_uncompressed)),
                        "utf-8",
                    )
                )
        except zlib.error as err:
            raise InvalidFileException(f"Unable to read table-of-contents: {err}")
        except OSError as err:
            raise FileAccessException(f"Unable to read archive: {err}")

    def _parse_entries(self, root, directory="") -> List[XAREntry]:
        """Recursively parse entries from the table-of-contents."""
        candidates = []

        # Strip any slashes, only using the last path component.
        kind = root.find(".type").text
        name = root.find(".name").text.split("/")[-1]
        path = os.path.join(directory, name)

        # Recurse for directories
        if kind == "directory":
            for element in root.findall(".//file"):
                candidates.extend(self._parse_entries(element, directory=path))

        if kind == "file":
            size = int(root.find(".//data/size").text)
            length = int(root.find(".//data/length").text)
            offset = int(root.find(".//data/offset").text)
            encoding = root.find(".//data/encoding").get("style")
            archived_cksum = root.find(".//data/archived-checksum").text
            archived_cksum_kind = root.find(".//data/archived-checksum").get("style")

            candidates.append(
                XAREntry(
                    length,
                    offset,
                    size,
                    encoding,
                    archived_cksum,
                    archived_cksum_kind,
                    path,
                    name,
                    kind,
                )
            )

        return candidates

    def entries(self) -> List[XAREntry]:
        """Return a list of entries in this XAR."""
        candidates = []

        for entry in self._toc.findall("./toc/file"):
            candidates.extend(self._parse_entries(entry))

        return candidates

    def extract(self, destination):
        """Extract all entries from the XAR to the optional destination directory."""
        # Offset must be adjusted by the size of the ToC and the header. This is as the
        # offset is from the first byte AFTER the header and compressed ToC.
        header_size = self._header.size + self._header.toc_length_compressed

        for entry in self.entries():
            parent = os.path.dirname(os.path.join(destination, entry.path))

            try:
                os.makedirs(parent, exist_ok=True)
            except OSError as err:
                raise FileAccessException(
                    f"Unable to create directory during extraction: {err}"
                )

            # Check whether a decompressor should be used.
            decompressor = None

            if entry.encoding == "application/x-gzip":
                decompressor = zlib.decompressobj(wbits=zlib.MAX_WBITS | 32).decompress

            # Perform extraction.
            # TODO: No decompression or integrity checking is performed today, nor are
            # ownership and modes followed.
            remaining = entry.length

            try:
                with open(self.archive, "rb") as fin:
                    with open(os.path.join(destination, entry.path), "wb") as fout:
                        fin.seek(header_size + entry.offset)

                        # Read all data in chunks to not balloon memory when processing
                        # large files.
                        while remaining > 0:
                            delta = remaining - CHUNK_SIZE
                            if delta < 0:
                                read_length = remaining
                            else:
                                read_length = CHUNK_SIZE

                            # Use a decompressor, if required.
                            if decompressor:
                                fout.write(decompressor(fin.read(read_length)))
                            else:
                                fout.write(fin.read(read_length))

                            remaining -= read_length
            except (OSError, zlib.error) as err:
                raise InvalidFileException(err)



import base64
from typing import List

from colorama import Fore, init
from stacs.scan import helper, model
from stacs.scan.__about__ import __version__
from stacs.scan.constants import ARCHIVE_FILE_SEPARATOR
from stacs.scan.model.finding import Sample


def generate_file_tree(virtual_path: str) -> str:
    """Returns a tree layout to the virtual path."""
    tree = str()
    parts = virtual_path.split(ARCHIVE_FILE_SEPARATOR)

    for index, part in enumerate(parts):
        # Add some style. Print a package / box before each archive, and a document
        # before the file.
        if (index + 1) == len(parts):
            emoji = "📄"
        else:
            emoji = "📦"

        tree += f"{' ' * (index * 4)}`-- {emoji} {part}\n"

    return tree.rstrip()


def generate_sample(sample: Sample):
    """Return a plain-text and text formatted sample."""
    # Ensure the sample is nicely base64 encoded if binary, rather than slapping three
    # already base64'd strings together.
    raw = bytearray()
    if sample.binary:
        raw.extend(bytearray(base64.b64decode(sample.before)))
        raw.extend(bytearray(base64.b64decode(sample.finding)))
        raw.extend(bytearray(base64.b64decode(sample.after)))

        return str(base64.b64encode(raw), "utf-8")

    return "".join([sample.before, sample.finding, sample.after])


def render(findings: List[model.finding.Entry], pack: model.pack.Format) -> str:
    """Render a 'pretty' output to the console for human consumption."""
    init()

    # Find all unsuppressed findings, and track them separately.
    results = {}
    unsuppressed = 0

    for finding in findings:
        # Check for suppressions.
        if finding.ignore is not None and finding.ignore.ignored:
            continue

        # Track it.
        unsuppressed += 1

        if results.get(finding.path) is None:
            results[finding.path] = []

        # Extract location appropriately.
        location = None
        if finding.location.line:
            location = f"line {finding.location.line}"
        else:
            location = f"{finding.location.offset}-bytes"

        # Generates all strings for presentation right away.
        results[finding.path].append(
            {
                "tree": generate_file_tree(finding.path),
                "path": finding.path.split(ARCHIVE_FILE_SEPARATOR)[-1],
                "rule": finding.source.reference,
                "text": finding.source.description,
                "location": location,
                "sample": generate_sample(finding.sample),
            }
        )

    # Provide a summary.
    print(helper.banner(version=__version__))

    if findings == 0:
        print("✨ " + Fore.GREEN + "No unsuppressed findings! Great work! ✨\n")
        return

    # Render out the findings.
    print(
        f"{Fore.RED}🔥 There were {unsuppressed} unsuppressed findings in "
        f"{len(results)} files 🔥\n"
    )

    for candidate in results:
        filepath = candidate.split(ARCHIVE_FILE_SEPARATOR)[0]
        count = len(results[candidate])

        if ARCHIVE_FILE_SEPARATOR in candidate:
            print(f"{Fore.RED}❌ {count} finding(s) inside of file {filepath} (Nested)")
        else:
            print(f"{Fore.RED}❌ {count} finding(s) inside of file {filepath}")

        for finding in results[candidate]:
            print()
            helper.printi(f"{Fore.YELLOW}Reason   : {finding['text']}")
            helper.printi(f"{Fore.YELLOW}Rule Id  : {finding['rule']}")
            helper.printi(f"{Fore.YELLOW}Location : {finding['location']}\n\n")
            helper.printi(f"{Fore.YELLOW}Filetree:\n\n")
            helper.printi(
                finding["tree"],
                prefix=f"    {Fore.RESET}|{Fore.BLUE}",
            )
            print()
            helper.printi(f"{Fore.YELLOW}Sample:\n\n")
            helper.printi(
                f"... {finding['sample']} ...",
                prefix=f"    {Fore.RESET}|{Fore.BLUE}",
            )
            print()

        print(f"\n{Fore.RESET}{'-' * 78}\n")
